/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define KERNEL
#include "ixemul.h"

#if __GNUC__ != 2
#define alloca __builtin_alloca
#endif

struct open_vec {
  int action;
  struct FileHandle *fh;
};

static int
__open_func (struct StandardPacket *sp, struct MsgPort *handler,
             BPTR parent_lock,
	     BSTR name,
	     struct open_vec *ov, int *no_error)
{
  sp->sp_Pkt.dp_Type = ov->action;
  sp->sp_Pkt.dp_Arg1 = CTOBPTR (ov->fh);
  sp->sp_Pkt.dp_Arg2 = parent_lock;
  sp->sp_Pkt.dp_Arg3 = name;

  PutPacket (handler, sp);
  __wait_packet (sp);

  *no_error = sp->sp_Pkt.dp_Res1 == -1;
 
  /* handler is supposed to supply this, but if it didn't... */
  if (*no_error && !ov->fh->fh_Type) ov->fh->fh_Type = handler;
 
  /* continue if we failed because of symlink - reference */
  return 1;
}

BPTR
__open (char *name, int action)
{
  unsigned long *fh;
  struct open_vec ov;
  BPTR res;

  /* provide a new filehandle, allocated as DOS-like as possible
   * (don't know if that's needed) */
  fh = (unsigned long *) kmemalign (4, sizeof (struct FileHandle) + 4);
  *fh = sizeof (struct FileHandle);
  ov.fh = (struct FileHandle *)(fh + 1);
  ov.action = action;
  bzero (ov.fh, sizeof (struct FileHandle));
  ov.fh->fh_Pos = ov.fh->fh_End = -1;

  /* special test for opening of NIL:. In that case, we just return this plain
   * initialized filehandle, that has a handler set to zero */
  if (! strcasecmp (name, "nil:"))
    res = -1;
  else
    res = __plock (name, __open_func, &ov);
  
  if (! res) kfree (fh);
  return res ? CTOBPTR (ov.fh) : 0;
}
